#include "declarations.h"
#include "memory.h"
#include "stack.h"
#include "heap.h"

/* =============Symbol Table =================================== */

_SymbolInfo  SymbolTable[1000];
int          numsym;


#define MAXLPOCACHE 400


int          lpoCache[MAXLPOCACHE][MAXLPOCACHE];
int LPOTimeStamp=1, LPOIndex=0;



short int STCheckSymbol(_Ident name, short int numargs)
Proc
Var
  short int i;
EndVar
  i=0;
  While i<numsym Do
    If strcmp( SymbolTable[i].name, name)==0 Then
      If SymbolTable[i].arity != numargs Then
        printf("\nError in arity of symbol %s\n\n",name);
      EndIf
      Return(i);
    Else
      i++;
    EndIf
  EndWhile
  strcpy(SymbolTable[i].name, name);
  SymbolTable[i].arity = numargs;
  numsym++;
  Return(i);
EndProc
void STPrint()
Proc
Var
  int i;
EndVar
  printf("\n\n Symbol Table: \n\n");
  For i=0;i<numsym;i++ Do
    If STIsAC(i) Then
      printf("%3u:  (AC)  %s with arity: %u\n",
           i,SymbolTable[i].name,SymbolTable[i].arity);
    Else
      printf("%3u:        %s with arity: %u\n",
           i,SymbolTable[i].name,SymbolTable[i].arity);
    EndIf
  EndFor
EndProc

_SymbolInfo  SymbolTable[1000];
int          numsym;



/* 
   numsym is the first empty position of the symbol table.
   The first "real" function symbol is number 4.
*/

void STInit()
Proc
  numsym = 0;
  STCheckSymbol("Ref",0);
  numsym = 1;
  STCheckSymbol("ErrorSym1",2);
  STCheckSymbol("ErrorSym2",2);
  STCheckSymbol("Eq",2);
EndProc


/* ============= Heap  =================================== */


_HeapAddr HeapCreateStruct(_HeapTag sym)
Proc
Var
  int j,arity;
  _HeapAddr addr;
EndVar
  arity = STArityOfSymbol(sym);
  addr = (_HeapAddr) MemoryAlloc( sizeof(_HeapCell)*(arity+1) );
  HeapSetTag(addr, sym);
  HeapSetLPOTimeStamp(addr, 0);
  HeapSetAddr(addr, (_HeapAddr) 1 );
  For j=1;j<=arity;j++ Do 
    HeapSetRef( HeapIthArg(addr,j) );
    HeapSetLPOTimeStamp(HeapIthArg(addr,j), 0);
  EndFor
  Return(addr);
EndProc


void HeapInitializeStruct(_HeapAddr addr, _HeapTag sym)
Proc
Var
  int j,arity;
EndVar
  arity = STArityOfSymbol(sym);
  HeapSetTag(addr, sym);
  HeapSetLPOTimeStamp(addr, 0);
  HeapSetAddr(addr, (_HeapAddr) 1 );
  For j=1;j<=arity;j++ Do 
    HeapSetRef( HeapIthArg(addr,j) );
    HeapSetLPOTimeStamp(HeapIthArg(addr,j), 0);
  EndFor
EndProc



void HeapFreeStruct(_HeapAddr addr)
Proc
Var
  int size,arity;
  _HeapTag sym;
EndVar
  size = sizeof(_HeapCell);
  If HeapIsRef(addr) Then
    arity = 0;
  Else
    sym = HeapTag(addr);
    arity = STArityOfSymbol(sym);
    size = size*(arity+1);
  EndIf
#if DEBUG
    {
    int j;
    HeapSetTag(addr,(_HeapTag)1);
    HeapSetAddr(addr,(_HeapAddr)0);
    For j=1;j<=arity;j++ Do 
      HeapSetTag(HeapIthArg(addr,j),(_HeapTag)1);
      HeapSetAddr(HeapIthArg(addr,j),(_HeapAddr)0);
    EndFor
    }
#endif
  MemoryFree((_Ptr)addr,size);
EndProc



void HeapFreeNormalTerm(_HeapAddr term)
Proc
Var
  int i, arity;
EndVar
  If HeapIsSymbol(term) Then
    arity = HeapTagArity(term);
    For i=1; i<=arity; i++ Do
      HeapFreeNormalTerm( HeapAddr(HeapIthArg(term,i)) );
    EndFor
    HeapFreeStruct(term);
  EndIf
EndProc


void HeapFreeTerm(_HeapAddr term)
Proc
Var
  int i, arity;
EndVar
  HeapDeref(term,term);
  If HeapIsSymbol(term) Then
    arity = HeapTagArity(term);
    For i=1; i<=arity; i++ Do
      HeapFreeTerm(HeapIthArg(term,i) );
    EndFor
    HeapFreeStruct(term);
  EndIf
EndProc


void HeapPrintRec( _HeapAddr addr, int i )
Proc
Var 
  int arity,j,k;
EndVar  
  For j=1;j<=i;j++ Do 
    printf("  ");
  EndFor
  printf("%u: | %s | %u | \n",
          (unsigned int) addr, 
          STNameOfSymbol(HeapTag(addr)),
          (unsigned int) HeapAddr(addr));
  If HeapIsRef(addr) And Not HeapIsSelfRef(addr) Then
    HeapPrintRec( HeapAddr(addr), i+1 );
  EndIf
  If Not HeapIsRef(addr) Then
    arity = HeapTagArity(addr);
    For k=1;k<=arity;k++ Do 
      For j=1;j<=i;j++ Do 
        printf("  ");
      EndFor
      printf("%u: | %s | %u | \n",
          (unsigned int) (addr+k), 
          STNameOfSymbol(HeapTag(addr+k)),
          (unsigned int) HeapAddr(addr+k));
      HeapPrintRec( HeapAddr(addr+k), i+1 );
    EndFor
  EndIf
EndProc



void HeapPrint( _HeapAddr addr )
Proc
  HeapPrintRec(addr,0); 
  printf("\n");
  printf("\n");
EndProc




void HeapWriteTerm(_HeapAddr term)
Proc
Var
  int i,arity;
  _HeapAddr dref;
EndVar
  If term == NULL Then
    printf("(NoTerm)");
  Else
    dref = term;
    HeapDeref(dref,dref);
    If HeapIsIntRef(dref) Then
      printf("I%p",dref);
    ElseIf HeapIsRef(dref) Then
      printf("X%d%d",VBBankNumber(dref),VBVarNumber(dref));
    Else
      arity = STArityOfSymbol( HeapTag( dref ));
      If HeapIsBlocked(dref) Then
        printf("#");
      EndIf
      printf("%s", STNameOfSymbol(HeapTag( dref )));
      If arity > 0 Then
        printf("(");
        For i=1; i<=arity; i++ Do
          HeapWriteTerm(HeapIthArg(dref,i));
          If i<arity Then
            printf(",");
          Else
            printf(")");
          EndIf
        EndFor
      EndIf
    EndIf
  EndIf
EndProc


void HeapWriteTermFile( FILE *f, _HeapAddr term)
Proc
Var
  int i,arity;
  _HeapAddr dref;
EndVar
  If term == NULL Then
    fprintf( f, "(NoTerm)");
  Else
    dref = term;
    HeapDeref(dref,dref);
    If HeapIsIntRef(dref) Then
      fprintf( f, "I%p",dref);
    ElseIf HeapIsRef(dref) Then
      fprintf( f, "X%d%d",VBBankNumber(dref),VBVarNumber(dref));
    Else
    /*    
      If HeapIsBlocked(dref) Then
        fprintf(f,"#");
      EndIf
      */ 
      arity = STArityOfSymbol( HeapTag( dref ));
      fprintf( f, "%s", STNameOfSymbol(HeapTag( dref )));
      If arity > 0 Then
        fprintf( f, "(");
        For i=1; i<=arity; i++ Do
          HeapWriteTermFile( f, HeapIthArg(dref,i));
          If i<arity Then
            fprintf( f, ",");
          Else
            fprintf( f, ")");
          EndIf
        EndFor
      EndIf
    EndIf
  EndIf
EndProc



/* copies arbitrary terms without changing the variable bank */

_HeapAddr HeapCopyTerm(_HeapAddr term)
Proc
Var
  int i,arity;
  _HeapAddr aux;
EndVar
  HeapDeref(term,term);
  If HeapIsRef(term) Then
    aux = term; 
  Else
    aux = HeapCreateStruct(HeapTag(term));
    arity = HeapTagArity(term);
    For i=1; i<=arity; i++ Do
      HeapSetAddr( HeapIthArg(aux,i), HeapCopyTerm(HeapIthArg(term,i)) );
    EndFor
  EndIf
  Return(aux);
EndProc


/* copies arbitrary terms without changing the variable bank */

_HeapAddr HeapCopyTermProp(_HeapAddr term)
Proc
Var
  int i,arity;
  _HeapAddr aux;
EndVar
  HeapDeref(term,term);
  If HeapIsRef(term) Then
    aux = term; 
  Else
    aux = HeapCreateStruct(HeapTag(term));
    If HeapIsBlocked(term) Then
      HeapSetBlocked(aux);
    EndIf
    arity = HeapTagArity(term);
    For i=1; i<=arity; i++ Do
      HeapSetAddr( HeapIthArg(aux,i), HeapCopyTermProp(HeapIthArg(term,i)) );
    EndFor
  EndIf
  Return(aux);
EndProc


_HeapAddr HeapCopyTermBlock(_HeapAddr term)
Proc
Var
  int i,arity;
  _HeapAddr aux;
  _Bool bank;
EndVar

  bank = FALSE;
  While HeapIsRefOrIntRef(term) And Not HeapIsSelfRef(term) Do
    If VBBankNumber(term) == bankInferences Or
       VBBankNumber(term) == bankNormal Then
      bank=TRUE;
    EndIf
    term = HeapAddr(term);
  EndWhile

  If HeapIsRef(term) Then
    aux = term; 
  Else
    aux = HeapCreateStruct(HeapTag(term));
    arity = HeapTagArity(term);
    For i=1; i<=arity; i++ Do
      HeapSetAddr( HeapIthArg(aux,i), HeapCopyTermBlock(HeapIthArg(term,i)) );
    EndFor
    If bank Then
      HeapSetBlocked(aux);
    EndIf
  EndIf
  Return(aux);
EndProc


/* copies normalized terms without changing the variable bank */
_HeapAddr HeapCopyNormalTerm(_HeapAddr term)
Proc
Var
  int i,arity;
  _HeapAddr aux;
EndVar
  If HeapIsRef(term) Then
    aux = term; 
  Else
    aux = HeapCreateStruct(HeapTag(term));
    arity = HeapTagArity(term);
    For i=1; i<=arity; i++ Do
      HeapSetAddr( HeapIthArg(aux,i), 
         HeapCopyTerm( HeapAddr( HeapIthArg(term,i))) );
    EndFor
  EndIf
  Return(aux);
EndProc




/* Equality of arbitrary terms */
_Bool HeapEqualTerms( _HeapAddr t1, _HeapAddr t2)
Proc
  Var
    int arity;
    _Bool match;
  EndVar
  HeapDeref(t1,t1);
  HeapDeref(t2,t2);
  If t1 == t2 Then
    Return(TRUE);
  ElseIf HeapIsRef(t1) Then
    Return(FALSE);
  ElseIf HeapTag(t1) == HeapTag(t2) Then
    arity = HeapTagArity(t1);
    match = TRUE;
    While arity > 0 And match Do
      match = HeapEqualTerms(HeapIthArg(t1,arity), HeapIthArg(t2,arity));
      arity--;
    EndWhile
    Return match;
  Else
    Return(FALSE);
  EndIf
EndProc


/* Equality of terms created with HeapCopyTerms */
_Bool HeapEqualNormalTerms( _HeapAddr t1, _HeapAddr t2)
Proc
  Var
    int arity;
    _Bool match;
  EndVar
  If t1 == t2 Then
    Return(TRUE);
  ElseIf HeapIsRef(t1) Then
    Return(FALSE);
  ElseIf HeapTag(t1) == HeapTag(t2) Then
    arity = HeapTagArity(t1);
    match = TRUE;
    While arity > 0 And match Do
      match = HeapEqualNormalTerms( HeapAddr(HeapIthArg(t1,arity)), 
                                    HeapAddr(HeapIthArg(t2,arity)) );
      arity--;
    EndWhile
    Return match;
  Else
    Return(FALSE);
  EndIf
EndProc


_HeapAddr HeapMoveToBankTerm(int *ctr, _HeapAddr term, short int bank)
Proc
Var
  int i, arity;
  _HeapAddr aux;
EndVar
  If HeapIsRef(term) Then
    If HeapIsSelfRef(term) Then
      aux = VBHeapAddr(bank,*ctr);
      (*ctr)++;
      HeapSetAddr(term, aux);
      StackPush( refStack, term );
      Return(aux);
    Else
      Return(HeapAddr(term));
    EndIf
  Else
    arity = HeapTagArity(term);
    For i=1;i<=arity;i++ Do
      HeapSetAddr( HeapIthArg(term,i), 
        HeapMoveToBankTerm(ctr, HeapAddr(HeapIthArg(term,i)), bank) );
    EndFor
    Return(term);
  EndIf
EndProc



void HeapTermRenameVars( short int bank, _HeapAddr term )
Proc
Var
  int arity;
  _HeapAddr ti, drefti;
EndVar
  arity = HeapTagArity(term);
  While arity Do
    ti = HeapIthArg(term,arity);
    drefti = HeapAddr(ti);
    If HeapIsRef(drefti) Then
      HeapSetAddr(ti, VBHeapAddr( bank, VBVarNumber(drefti) ) );
    Else
      HeapTermRenameVars(bank,drefti);
    EndIf
    arity--;
  EndWhile
EndProc


/* ============= Variable Banks  =================================== */

_VarBank     varbanks[3];

void VBInit()
Proc
Var 
  int i,j;
EndVar
  For i=0;i<3;i++ Do
    For j=0;j<200;j++ Do
      HeapSetRef( VBHeapAddr(i,j) );
      HeapSetSelfRef( VBHeapAddr(i,j) );
    EndFor
  EndFor
EndProc


/* ============= End Heap.c  =================================== */


_Bool LPOGr(_HeapAddr s, _HeapAddr t, _Bool ct);
_Bool LPOGrArgsStarting(_HeapAddr s, _HeapAddr t, int first, _Bool ct);
_Bool LPOGrLex(_HeapAddr s, _HeapAddr t, _Bool ct);
_Bool LPOTermContainsVar(_HeapAddr s, _HeapAddr v);
_Bool LPOSomeArgGEQ(_HeapAddr s, _HeapAddr t, _Bool ct);


int TermSize(_HeapAddr t, int var[200], int *max)
Proc
Var
  int arity, n, num;
EndVar
  If HeapIsRef(t) Then
    num = VBVarNumber(t);
    If num > *max Then
      *max = num;
    EndIf 
    var[num]++;
    Return(1);
  Else
    n=1;
    arity = HeapTagArity(t);
    While arity>0 Do
      n += TermSize( HeapAddr( HeapIthArg(t,arity) ), var, max);
      arity--;
    EndWhile
  EndIf
  Return n;
EndProc



int vars[200], vart[200];

void LPOInit()
Proc
Var
  int i;
EndVar
  For i=0;i<200;i++ Do
    vars[i]=0;
    vart[i]=0;
  EndFor
EndProc


_Bool SizeOrient(_HeapAddr s, _HeapAddr t, _HeapAddr *l, _HeapAddr *r)
Proc
Var
  int  sizes, sizet, max, i;
  _Bool check;
EndVar
  max = -1;
  sizes = TermSize(s,vars,&max);
  sizet = TermSize(t,vart,&max);
  If sizes>sizet Then
    check = TRUE;
    i=0;
    While check And i<=max Do
      If vars[i]<vart[i] Then
        check=FALSE;
      EndIf
      i++;
    EndWhile
    *l = s;
    *r = t;
  ElseIf sizet>sizes Then
    check = TRUE;
    i=0;
    While check And i<=max Do
      If vart[i]<vars[i] Then
        check=FALSE;
      EndIf
      i++;
    EndWhile
    *l = t;
    *r = s;
  Else
    check = TRUE;
    i=0;
    If LPOGr(s,t,FALSE) Then
      While check And i<=max Do
        If vars[i]<vart[i] Then
          check=FALSE;
        EndIf
        i++;
      EndWhile
      *l = s;
      *r = t;
    ElseIf LPOGr(t,s,FALSE) Then
      While check And i<=max Do
        If vart[i]<vars[i] Then
          check=FALSE;
        EndIf
        i++;
      EndWhile
      *l = t;
      *r = s;
    Else
      *l = s;
      *r = t;
      check=FALSE;
    EndIf
  EndIf
  For i=0;i<=max;i++ Do
    vars[i]=0;
    vart[i]=0;
  EndFor
  Return check;
EndProc



_Bool LPOOrient(_HeapAddr s, _HeapAddr t, _HeapAddr *l, _HeapAddr *r)
Proc

  LPOCacheInit();

  If LPOGr(s,t,0) Then
    *l = s;
    *r = t;
    Return TRUE;
  EndIf

  LPOCacheInit();

  If LPOGr(t,s,0) Then
    *l = t;
    *r = s;
    Return TRUE;
  EndIf

  *l = s;
  *r = t;
  Return FALSE;

EndProc


_Bool TermGreaterTopSymbolPrecedence(_HeapAddr s, _HeapAddr t)
Proc
/*
 Return(HeapTag(s)>HeapTag(t));
 */

  If HeapTagArity(s) > HeapTagArity(t) Then
    Return(TRUE);
  ElseIf HeapTagArity(s) == HeapTagArity(t) Then
    Return( HeapTag(s) < HeapTag(t) );
  Else
    Return(FALSE);
  EndIf

EndProc


void LPOCacheInit()
Proc
  LPOTimeStamp++;
  LPOIndex =0;
EndProc


_Bool LPOGr(_HeapAddr s, _HeapAddr t, _Bool ct)
Proc
Var
  _Bool bigger;
EndVar

  HeapDeref(s,s);
  HeapDeref(t,t);

If HeapLPOTimeStamp(s) == LPOTimeStamp And
   HeapLPOTimeStamp(t) == LPOTimeStamp And
   (lpoCache[HeapLPOIndex(s)][HeapLPOIndex(t)] ==  LPOTimeStamp Or
    lpoCache[HeapLPOIndex(s)][HeapLPOIndex(t)] == -LPOTimeStamp) Then
  bigger = lpoCache[HeapLPOIndex(s)][HeapLPOIndex(t)]>0;
  Return(bigger);
EndIf


If HeapIsRef(s) Then
  bigger=FALSE;
ElseIf HeapIsRef(t) Then
/*  
  If ct And VBBankNumber(t) ==  bankDemod Then
    StackPush(demodRefStack,t);
    HeapSetAddr(t, smallestct );
    t=smallestct;
    bigger=LPOGr(s,t,ct);
  Else
    bigger=TermGreaterTopSymbolPrecedence(s,t);
  EndIf 
*/
  bigger = LPOTermContainsVar(s,t);
ElseIf TermGreaterTopSymbolPrecedence(s,t) Then
  bigger = LPOGrArgsStarting(s,t,1,ct); 
ElseIf HeapTag(s) == HeapTag(t) Then
  bigger = LPOGrLex(s,t,ct);
Else
  bigger = LPOSomeArgGEQ(s,t,ct);
EndIf

If HeapLPOTimeStamp(s) != LPOTimeStamp Then
  HeapSetLPOTimeStamp(s,LPOTimeStamp);
  HeapSetLPOIndex(s,LPOIndex);
  LPOIndex++;
  If LPOIndex>=MAXLPOCACHE Then
    printf("error LPOCACHE overflow!\n"); exit(1);
  EndIf
EndIf

If HeapLPOTimeStamp(t) != LPOTimeStamp Then
  HeapSetLPOTimeStamp(t,LPOTimeStamp);
  HeapSetLPOIndex(t,LPOIndex);
  LPOIndex++;
  If LPOIndex>=MAXLPOCACHE Then
    printf("error LPOCACHE overflow!\n"); exit(1);
  EndIf
EndIf

If bigger Then 
  lpoCache[HeapLPOIndex(s)][HeapLPOIndex(t)]= LPOTimeStamp;
  lpoCache[HeapLPOIndex(t)][HeapLPOIndex(s)]=-LPOTimeStamp;
Else
  lpoCache[HeapLPOIndex(s)][HeapLPOIndex(t)]=-LPOTimeStamp;
EndIf

Return(bigger);
EndProc



_Bool LPOTermContainsVar(_HeapAddr s, _HeapAddr v)
Proc
Var
  int arity;
  _Bool contains;
  _HeapAddr s1;
EndVar
arity=HeapTagArity(s);
contains=FALSE;
While Not contains And arity Do
  s1 = HeapAddr(HeapIthArg(s,arity));
  HeapDeref(s1,s1);
  If HeapIsRef(s1) Then
    contains = ( s1 == v );
  Else
    contains = LPOTermContainsVar(s1,v);
  EndIf
  arity--;
EndWhile
Return(contains);
EndProc


_Bool LPOSomeArgGEQ(_HeapAddr s, _HeapAddr t, _Bool ct)
Proc
Var
  int arity;
  _Bool geq;
  _HeapAddr si;
EndVar
arity=HeapTagArity(s);
geq=FALSE;
While Not geq And arity Do
  si = HeapAddr( HeapIthArg(s,arity) );
  geq = ( LPOGr(si,t,ct) Or HeapEqualTerms(si,t) );
  arity--;
EndWhile
Return(geq);
EndProc


_Bool LPOGrArgsStarting(_HeapAddr s, _HeapAddr t, int first, _Bool ct)
Proc
Var
  _HeapAddr ti;
  int arity;
  _Bool greater;
EndVar
arity=HeapTagArity(t);
greater=TRUE;
While greater And first<=arity Do
  ti = HeapAddr(HeapIthArg(t,first));
  greater = LPOGr(s,ti,ct);
  first++;
EndWhile
Return(greater);
EndProc


_Bool LPOGrLex(_HeapAddr s, _HeapAddr t, _Bool ct)
Proc
Var
  _HeapAddr si;
  int i,arity;
  _Bool equal,geq;
EndVar
i=1;
arity=HeapTagArity(s);
equal = TRUE;
While equal And i<=arity Do
  equal = HeapEqualTerms( HeapAddr(HeapIthArg(s,i)),
                          HeapAddr(HeapIthArg(t,i)) );
  i++;
EndWhile
If equal Then  /* exitted while because i>arity: all args equal */
  Return(FALSE);
Else
  If LPOGr( HeapAddr(HeapIthArg(s,i-1)), 
            HeapAddr(HeapIthArg(t,i-1)), ct ) Then 
    Return( LPOGrArgsStarting(s,t,i,ct) );
  Else
    geq=FALSE;
    While Not geq And i<=arity Do
      si = HeapAddr( HeapIthArg(s,i) );
      geq = ( LPOGr(si,t,ct) Or HeapEqualTerms(si,t) );
      i++;
    EndWhile
    Return(geq);
  EndIf
EndIf
EndProc


