#ifndef HEAP_H
#define HEAP_H

#include "memory.h"

/* ================ Symbol Table: ===================================== */

typedef char _Ident[50];

typedef struct
  {
  _Ident      name;
  short int   arity;
  _Bool       isac;
  } _SymbolInfo;

extern _SymbolInfo  SymbolTable[1000];

extern int          numsym;

#define STArityOfSymbol(i) (SymbolTable[i].arity)
#define STNameOfSymbol(i)  (SymbolTable[i].name)
#define STSetAC(i)         (SymbolTable[i].isac=TRUE)
#define STIsAC(i)         (SymbolTable[i].isac)



extern short int STCheckSymbol(_Ident name, short int numargs);
extern void STPrint();
extern void STInit();


/* ================ Type definitions:  ==================================== */

typedef short int _HeapTag;

struct _HEAPCELL;
typedef struct _HEAPCELL _HeapCell;

struct _HEAPCELL
           {
           _HeapTag    tag;
           _HeapCell*  address;
           int         lpoind,lpots;
           };

typedef _HeapCell *_HeapAddr;
typedef _HeapCell _VarBank[200];

extern _VarBank     varbanks[3];
extern _HeapAddr    inputLeftTerms[1000],inputRightTerms[1000];
extern int          numterms;

#define bankNormal         0
#define bankDemod          1
#define bankInferences     2

#define iow(term)   HeapWriteTerm(term);


/* ================ Macros:  ============================================ */
#define    HeapLPOIndex(addr)             ( (addr)->lpoind )
#define    HeapSetLPOIndex(addr, val)     ( (addr)->lpoind = (val) )
#define    HeapLPOTimeStamp(addr)         ( (addr)->lpots )
#define    HeapSetLPOTimeStamp(addr, val) ( (addr)->lpots = (val) )

#define    HeapTag(addr)            ( (addr)->tag )
#define    HeapTagArity(addr)       ( STArityOfSymbol((addr)->tag) )
#define    HeapAddr(addr)           ( ((addr)->address) )
#define    HeapSetTag(addr, val)    ( (addr)->tag = (val) )
#define    HeapSetRef(addr)         ( (addr)->tag = 0 )
#define    HeapSetIntRef(addr)      ( (addr)->tag = -1 )
#define    HeapSetAddr(addr, val)   ( (addr)->address = (val) )
#define    HeapIthArg(addr, i)      ( (addr) + (i) )
#define    HeapIsSymbol(addr)       ( HeapTag(addr) > 0 )
#define    HeapIsRef(addr)          ( HeapTag(addr) == 0 )
#define    HeapIsIntRef(addr)       ( HeapTag(addr) < 0 )
#define    HeapIsRefOrIntRef(addr)  ( HeapTag(addr) <= 0 )
#define    HeapIsSelfRef(addr)      ( (addr) == HeapAddr(addr) )
#define    HeapSetSelfRef(addr)     ( HeapSetAddr((addr), (addr)) )

#define    HeapDeref(addr, dref)    { \
  dref = addr; \
  While HeapIsRefOrIntRef(dref) And Not HeapIsSelfRef(dref) Do \
    dref = HeapAddr(dref); \
  EndWhile }

#define    HeapDerefInt(addr, dref)    { \
  dref = addr; \
  While HeapIsRef(dref) And Not HeapIsSelfRef(dref) Do \
    dref = HeapAddr(dref); \
  EndWhile }

#define    HeapCreateRef(addr)      { \
  addr = (_HeapAddr) MemoryAlloc(sizeof(_HeapCell)); \
  HeapSetRef(addr); }

#define    HeapCreateIntRef(addr)   { \
  addr = (_HeapAddr) MemoryAlloc(sizeof(_HeapCell)); \
  HeapSetIntRef(addr); } 


#define VBHeapAddr(banknum, numvar)  (&(varbanks[banknum][numvar]))
#define VBBankNumber(add)            ( (add - &(varbanks[0][0]) ) Div 200 )
#define VBVarNumber(add)             ( (add - &(varbanks[0][0]) ) Mod 200 )



#define HeapIsNormalForm(t)    (((int) ((t)->address)) == -nftimestamp  \
                             Or ((int) ((t)->address)) ==  nftimestamp  )
#define HeapSetNormalForm(t)    {                         \
       If ((int) ((t)->address))<0 Then                   \
          ((int) ((t)->address)) = -nftimestamp;          \
       Else                                               \
          ((int) ((t)->address)) =  nftimestamp;          \
       EndIf }


#define HeapIsBlocked(t)  ( ((int)((t)->address))<0 )

#define HeapSetBlocked(t)      {                            \
       If ((int) ((t)->address))>0 Then                     \
          ((int) ((t)->address)) = -((int) ((t)->address)); \
       EndIf }

#define HeapSetUnBlocked(t)    {                            \
       If ((int) ((t)->address))<0 Then                     \
          ((int) ((t)->address)) = -((int) ((t)->address)); \
       EndIf }


extern _HeapAddr HeapCreateStruct(_HeapTag sym);
extern void      HeapInitializeStruct(_HeapAddr addr, _HeapTag sym);
extern void      HeapFreeStruct(_HeapAddr addr);
extern void      HeapFreeNormalTerm(_HeapAddr addr);
extern void      HeapFreeTerm(_HeapAddr addr);
extern void      HeapPrint( _HeapAddr addr );
extern void      HeapWriteTerm(_HeapAddr term);
extern void      HeapWriteTermFile(FILE *f, _HeapAddr term);
extern void      VBInit();
extern void      LPOInit();
extern _Bool     HeapEqualTerms(_HeapAddr term1, _HeapAddr term2);
extern _Bool     HeapEqualNormalTerms(_HeapAddr term1, _HeapAddr term2);
extern _HeapAddr HeapCopyTerm( _HeapAddr term);
extern _HeapAddr HeapCopyTermProp( _HeapAddr term);
extern _HeapAddr HeapCopyTermBlock( _HeapAddr term);
extern _HeapAddr HeapCopyNormalTerm( _HeapAddr term);
extern _HeapAddr HeapMoveToBankTerm(int *ctr, _HeapAddr term, short int bank);


extern _Bool     LPOOrient(_HeapAddr s, _HeapAddr t, _HeapAddr *l, _HeapAddr *r);
extern _Bool     LPOGr(_HeapAddr s, _HeapAddr t, _Bool ct);
extern void      LPOCacheInit();
extern _Bool     SizeOrient(_HeapAddr s, _HeapAddr t, _HeapAddr *l, _HeapAddr *r);
extern void      HeapTermRenameVars( short int bank, _HeapAddr term );

#endif
