#include "stack.h"
#include "stree.h"
#include "flags.h"
#include "clock.h"
#include "main.h"


#define BASIC      1      /* basicness on/off                                */
#define NOFWD      1      /*  forward demod w/ non-orientable rules on/off   */
#define NOBWD      1      /* backward demod w/ non-orientable rules on/off   */
#define INILIM     27     /* initial limit */
#define CPLIMCOEF      1.5    /*     cp   removed when size > limit * x      */
#define CPTHMLIMCOEF   3.2    /* thm cp   removed when size > limit * x      */
#define RULELIMCOEF    1      /* rule     removed when size > limit * x      */
#define RULETHMLIMCOEF 1.5    /* thm rule removed when size > limit * x      */
#define INIINC     4      /* when numrules > inc*1000, exit inner loop       */
#define INCLIMIT   4      /* each time inner loop ends, increment limit w/ x */
#define INCINC     4      /* each time inner loop ends, increment inc   w/ x */



#define CPDEBUG 1
#define RWDEBUG 1


#define REWRITEHEAPSIZE 800000
#define MAXNUMCPS 10000

#if RWDEBUG
#define CheckRewriteHeapFFree  ({ \
 If rewriteHeapFFree >= REWRITEHEAPSIZE Then \
   printf("rewriteHeapFFree overflow!\n\n"); exit(1); \
 EndIf })
#else
#define CheckRewriteHeapFFree
#endif


#if CPDEBUG
#define CheckNumCPS  ({ \
 If numcps >= MAXNUMCPS Then \
   printf("numcps overflow!\n\n"); exit(1); \
 EndIf })
#else
#define CheckNumCPS
#endif


/*-----------------------------------------Local procedure headers */

void         AddNewRule( _HeapAddr l, _HeapAddr r, int p1, int p2 );
_HeapAddr    InsertSharingSub(_HeapAddr term, int num);
void         RemoveSharingSub( _HeapAddr term );
void         PrintRules( void );
void         SizeInit( void );
int          GetOldestNewRuleNum();
int          GetSmallestNewRuleNum( void );
int          GetSmallestThmNewRuleNum( void );
void         AddNewRuleSize( int size, int rulenum );
void         AddNewThmRuleSize( int size, int rulenum );
int          ComputeRuleSize( _HeapAddr l, _HeapAddr r );
int          ComputeTermSize( _HeapAddr t );
void         CritPairs( int n);
void         ComputeNormalForm( _HeapAddr s );
void         StoreCP( _HeapAddr s, _HeapAddr t, int p1, int p2 );
void         FreeCP();
void         TreatCP();
void         BacktrackRewriteStack( void );
_Bool        IsGround(_HeapAddr th1);
_Bool        Subsumed(_HeapAddr ns, _HeapAddr nt);
void         PrintProof(int n);
void         RuleSimplification( int i);
_HeapAddr    NormalFormlhs( _HeapAddr s );
_HeapAddr    NormalFormrhs( _HeapAddr s );
_IntListNode *GetListDemodSteps();
void         RecursivelySimplify( int before );
void         RuleCopyToBank(short int bank, _HeapAddr   l, _HeapAddr   r, 
                                    _HeapAddr *nl, _HeapAddr *nr);

/*--------------------global variables this module-------------------------*/
Var
  _STree       demod, ndemod, sub;
  _IntListNode *rulesizes[MAXRULESIZE];
  _IntListNode *thmrulesizes[MAXRULESIZE];
  int          smallestsize, thmsmallestsize, oldestnewrule=0;
  int          numSimplified, simplified[30000];
  int          numnewrules=0, numthmnewrules=0, numrules=0;
  _RuleNode    *rules[MAXNUMRULES];
  _HeapAddr    refTerm, refTermlhs, refTermrhs;
  _HeapAddr    inputLeftTerms[1000],inputRightTerms[1000];
  _HeapAddr    equation, th1,th2,th1r,th2r;
  int          numterms=1, nftimestamp=2;

  int          rewriteHeapFFree=0;
  _HeapCell    rewriteHeap[REWRITEHEAPSIZE];

  int          numcps=0;
  _CPStruct    critpairs[MAXNUMCPS];

  int          limit;
  _Stack       demodRefStack, refStack, rewritestack;
  _Stack       demodsteps;
  _HeapTag     thmtrue,thmfalse,theoremtop;
EndVar


/*--------------------main program-----------------------------*/

void main()
Proc
Var
  int          inc,i, cycles;
EndVar
  StackFlush(      refStack );
  StackFlush( demodRefStack );
  StackFlush( rewritestack  );
  Clock_Reset;
  Clock_Start;
  STInit();
  VBInit();
  LPOInit();  /* only if non-LPO ordering is used */
  STreeInitStacks();
  MemoryInit();
  SizeInit();
  HeapCreateRef(refTerm);    HeapSetSelfRef(refTerm);
  HeapCreateRef(refTermlhs); HeapSetSelfRef(refTermlhs);
  HeapCreateRef(refTermrhs); HeapSetSelfRef(refTermrhs);
  equation =  HeapCreateStruct( STCheckSymbol("$eq$", 2) );

/*----------------------------------------INPUT */
  yyparse();

/* STPrint(); */

  flag3(7,inputLeftTerms[0],inputRightTerms[0]);
  theoremtop    = STCheckSymbol("thtop", 2);
  thmtrue       = STCheckSymbol("thmtrue",  0);
  thmfalse      = STCheckSymbol("thmfalse",  0);
  th1 =  HeapCreateStruct( theoremtop );
  th2 =  HeapCreateStruct( theoremtop );
  th1r = HeapCreateStruct( thmtrue );
  th2r = HeapCreateStruct( thmfalse );
  HeapSetAddr( HeapIthArg(th1,1), inputLeftTerms[0] );
  HeapSetAddr( HeapIthArg(th1,2), inputRightTerms[0] );
  HeapSetAddr( HeapIthArg(th2,1), VBHeapAddr(bankDemod, 0) );
  HeapSetAddr( HeapIthArg(th2,2), VBHeapAddr(bankDemod, 0) );
  limit=10000;
  StoreCP(th1,th1r,0,0);
  StoreCP(th2,th2r,0,0);
  For i=1;i<numterms;i++ Do
    StoreCP( inputLeftTerms[i], inputRightTerms[i], 0,0);
  EndFor

  While numcps > 0 Do
    TreatCP();
  EndWhile


/*----------------------------------------COMPLETION PROCEDURE */
cycles = 0;
limit=INILIM;
inc= INIINC;

While numrules<100000 Do
  While (numnewrules > 0) And (numrules<1000*inc) Do
     cycles++;
      If (cycles Mod 3 == 0 And numthmnewrules > 0 ) Then
        i = GetSmallestThmNewRuleNum();
      ElseIf (cycles Mod 3) == 0 Then
        i = GetOldestNewRuleNum();
      Else
        i = GetSmallestNewRuleNum();
      EndIf


    flag4(2,(_HeapAddr)i, rules[i]->lhs, rules[i]->rhs);
    numnewrules--;
    If HeapTag(rules[i]->lhs)==theoremtop Then
      numthmnewrules--;
    EndIf
    rules[i]->isold = TRUE;
    CritPairs(i);
    While numcps > 0 Do
      TreatCP(); 
    EndWhile
  EndWhile

  inc = inc+INIINC;
  limit = limit+INCLIMIT;
  oldestnewrule = 0;
/* printf("L%d.\n",limit); */
  For i=1; i<=numrules; i++ Do
    If Not DeletedRule(i) And rules[i]->isold Then
      numnewrules++;
      rules[i]->isold=FALSE;
      If HeapTag(rules[i]->lhs)==theoremtop Then
        numthmnewrules++;
        AddNewThmRuleSize( rules[i]->size, i );
      EndIf        
      AddNewRuleSize( rules[i]->size, i );
    EndIf
  EndFor

EndWhile
    /* PrintRules();  */
EndProc

/*---------------------------------------- END MAIN  */


void TreatCP()
Proc
Var
  _Bool thm;
  _HeapAddr s,t,ns,nt,nns,nnt;
  int p1,p2, size;
EndVar


  p1 = critpairs[numcps].p1;
  p2 = critpairs[numcps].p2;
  s = critpairs[numcps].lhs;
  t = critpairs[numcps].rhs;

  thm  = HeapTag(s)==theoremtop Or HeapTag(t)==theoremtop;
  size = ComputeTermSize(s) + ComputeTermSize(t);

  flag3(8,s,t);
  StackFlush( demodsteps ); 

  If p2!=0 And !thm And size > limit * CPLIMCOEF Then
    FreeCP();
  ElseIf p2!=0 And thm And size > limit * CPTHMLIMCOEF Then
    FreeCP();
  Else
    nftimestamp++;
    ns = NormalFormlhs(s);
    nt = NormalFormrhs(t);

    size = ComputeTermSize(ns) + ComputeTermSize(nt);

    If (p2 != 0) And !thm And size > limit * RULELIMCOEF Then
      BacktrackRewriteStack();
      FreeCP();
    ElseIf (p2 != 0) And thm And size > limit * RULETHMLIMCOEF Then
      BacktrackRewriteStack();
      FreeCP();
    ElseIf HeapEqualTerms(ns,nt) Then
      flag1(1);
      BacktrackRewriteStack();
      FreeCP();
    ElseIf Subsumed(ns,nt) Then
      flag1(2);
      BacktrackRewriteStack();
      FreeCP();
    Else
      RuleCopyToBank(bankDemod,ns,nt,&nns,&nnt);
      BacktrackRewriteStack();
      FreeCP();
      AddNewRule( nns, nnt, p1, p2 );
    EndIf
  EndIf
EndProc


_Bool Variants(_HeapAddr s, _HeapAddr t)
Proc
Var
  _Bool result;
  _HeapAddr u,v;
EndVar
  RuleCopyToBank(bankNormal,s,t,&u,&v);
  result = TermMatch(u,t);
  RestoreRefsUntilEmpty(demodRefStack);
  If result Then
    result = TermMatch(v,s);
    RestoreRefsUntilEmpty( demodRefStack );    
  EndIf
  HeapFreeTerm(u);
  HeapFreeTerm(v);
  Return(result);
EndProc  


void AddNewRule( _HeapAddr s, _HeapAddr t, int p1, int p2 )
Proc
Var 
  int *info, size;
  _HeapAddr l, r;
  _Bool proved;
EndVar
  numnewrules++;
  numrules++;
  proved = FALSE;

/*
  If (numrules Mod 1000)==0 Then
    printf("R%d\n",numrules Div 1000);
  EndIf
*/

  rules[numrules] = (_RuleNode *) MemoryAlloc(sizeof(_RuleNode));

  If LPOOrient(s,t,&l,&r) Then
    size = 2*ComputeTermSize(l)+ComputeTermSize(r);
    rules[numrules]->oriented  = TRUE;
    flag4(1,(_HeapAddr)numrules,l,r);
    info = (int *) STreeInsert(&demod,l);
    *info = numrules;
    If HeapTag(l) == theoremtop Then
      numthmnewrules++;
      AddNewThmRuleSize( size, numrules );
    EndIf
  Else
    size = 2 * (ComputeTermSize(l) + ComputeTermSize(r));
    rules[numrules]->oriented  = FALSE;
    flag4(5,(_HeapAddr)numrules,l,r);
    If HeapIsRef(l) Or HeapIsRef(r) Then
      proved = TRUE;
    ElseIf NOFWD Then
      info = (int *) STreeInsert(&demod,l);
      If !(*info) Then
        *info = numrules;
      EndIf
      If !Variants(l,r) Then
        info = (int *) STreeInsert(&demod,r);
        If !(*info) Then
          *info = -numrules;
        EndIf
      EndIf
    EndIf

    HeapSetAddr( HeapIthArg(equation,1),l);
    HeapSetAddr( HeapIthArg(equation,2),r);
    info = (int *) STreeInsert(&ndemod,equation);
    *info = numrules;
  EndIf
  AddNewRuleSize( size, numrules );

  rules[numrules]->parent1   = p1;
  rules[numrules]->parent2   = p2;
  rules[numrules]->demodlist = GetListDemodSteps();
  rules[numrules]->isold     = FALSE;
  rules[numrules]->isdeleted = FALSE;
  rules[numrules]->size      = size;
  rules[numrules]->lhs       = l;
  rules[numrules]->rhs       = r;
  
  If HeapTag(l)==thmtrue And HeapTag(r)==thmfalse Then
    proved = TRUE;
  ElseIf HeapTag(r)==thmtrue And
         TermUnify( HeapIthArg(l,1),HeapIthArg(l,2) ) Then
    proved = TRUE;
  EndIf

  RestoreRefsUntilEmpty( refStack );

  If proved Then
    printf("Theorem Proved.  "); 
 /* PrintRules(); */
    PrintProof(numrules);
/*   Clock_Stop; Clock_Print;*/ MemoryStatistics();  printf("\n");
    exit(1);
  EndIf

  If NOBWD Or rules[numrules]->oriented Then
    RuleSimplification(numrules);
  EndIf

  If HeapTag(r)!=thmtrue Then
    InsertSharingSub(rules[numrules]->lhs, numrules);
    If HeapIsSymbol(rules[numrules]->rhs) Then 
      InsertSharingSub(rules[numrules]->rhs, numrules);
    EndIf
  EndIf
EndProc


void RemoveBlocksFromTerm(_HeapAddr term)
Proc
Var
  int i,arity;
EndVar
  HeapDeref(term,term);
  If Not HeapIsRef(term) Then
    HeapSetUnBlocked(term);
    arity = HeapTagArity(term);
    For i=1; i<=arity; i++ Do
      RemoveBlocksFromTerm( HeapIthArg(term,i) );
    EndFor
  EndIf
EndProc


void RuleSimplification(int i)
Proc
Var
  _IntListNode *p,*p1;
  _InfoNodeSub *infoSub;
  int hit,j;
  _Bool oriented;
  _HeapAddr l,r;
EndVar

  numSimplified = 0;
  flag2(7,(_HeapAddr)i);
  RuleCopyToBank( bankNormal, rules[i]->lhs, rules[i]->rhs, &l, &r);
  oriented = rules[i]->oriented;


  infoSub = STreeFirstReverseMatch( sub, l );
  While infoSub Do
    p = infoSub->numslist;
    While p Do
      hit = p->num;
      LPOCacheInit();
      If hit != i And Not DeletedRule(hit) And (oriented Or LPOGr(l,r,0)) Then
        flag2(8,(_HeapAddr)hit);
        numSimplified++;
        simplified[numSimplified] = hit;
      EndIf
      While p->next And DeletedRule( ((_IntListNode *)p->next)->num) Do
        p1 = p->next;
        p->next = ((_IntListNode *)p->next)->next;
        MemoryFree((_Ptr) p1, sizeof(_IntListNode));
      EndWhile
      p = p->next;
    EndWhile
    infoSub = STreeNextReverseMatch();
  EndWhile

  If !oriented And !Variants(l,r) Then
    infoSub = STreeFirstReverseMatch( sub, r );
    While infoSub Do
      p = infoSub->numslist;
      While p Do
        hit = p->num;
        LPOCacheInit();
        If hit != i And Not DeletedRule(hit) And LPOGr(r,l,0) Then
          flag2(8,(_HeapAddr)hit);
          numSimplified++;
          simplified[numSimplified] = hit;
        EndIf
        While p->next And DeletedRule( ((_IntListNode *)p->next)->num) Do
          p1 = p->next;
          p->next = ((_IntListNode *)p->next)->next;
          MemoryFree((_Ptr) p1, sizeof(_IntListNode));
        EndWhile
        p = p->next;
      EndWhile
      infoSub = STreeNextReverseMatch();
    EndWhile
  EndIf

  HeapFreeNormalTerm(l);
  HeapFreeNormalTerm(r);

  /* con esto comentado ROB004-1 se resiente: */

  If numSimplified != 0 Then
    RemoveBlocksFromTerm(rules[i]->lhs);
    RemoveBlocksFromTerm(rules[i]->rhs);
  EndIf

  While numSimplified != 0 Do
    j = simplified[numSimplified--];
    If Not DeletedRule(j) Then
      If rules[j]->oriented Then
        STreeDelete( &demod, rules[j]->lhs );
      Else
        HeapSetAddr( HeapIthArg(equation,1), rules[j]->lhs );
        HeapSetAddr( HeapIthArg(equation,2), rules[j]->rhs );
        STreeDelete( &ndemod, equation );
      EndIf

      StoreCP( rules[j]->lhs, rules[j]->rhs, j, 0 );
      If Not rules[j]->isold Then
        numnewrules--;
        If HeapTag(rules[j]->lhs)==theoremtop Then
          numthmnewrules--;
        EndIf
      EndIf
      RemoveSharingSub(rules[j]->lhs);
      If HeapIsSymbol(rules[j]->rhs) Then 
        RemoveSharingSub(rules[j]->rhs);
      EndIf
      rules[j]->isdeleted = TRUE;
      flag2(6,(_HeapAddr)j);
    EndIf
  EndWhile
  flag2(10,(_HeapAddr)i);
EndProc



void AddNumToSubTerms( _HeapAddr term, int num )
Proc
Var
  _IntListNode *p;
  int arity;
  _HeapAddr ti;
  _InfoNodeSub *node;
EndVar

  node = (_InfoNodeSub*) 
     ((unsigned int)term - sizeof(_IntListNode*) - sizeof(int)); 
  If (node->numslist)->num != num Then
    p = (_IntListNode *) MemoryAlloc(sizeof(_IntListNode));
    p->num = num;
    p->next = node->numslist;
    node->numslist = p;
  EndIf
  (node->ctr)++;
  arity = HeapTagArity(term);
  While arity > 0 Do
    ti = HeapAddr( HeapIthArg(term,arity));
    If Not HeapIsSelfRef(ti) Then
      AddNumToSubTerms( ti, num);
    EndIf
    arity--;
  EndWhile
EndProc


_HeapAddr InsertSharingSub(_HeapAddr term, int num)
Proc
Var
  _HeapTag sym;
  _HeapAddr ti, newterm;
  _InfoNodeSub **info; 
  _IntListNode *p;
  int arity;
EndVar

  info = (_InfoNodeSub **) STreeFind(sub,term);
  If info Then
    newterm = &((*info)->cell);
    AddNumToSubTerms(newterm, num);
  Else
    info = (_InfoNodeSub **) STreeInsert(&sub,term);
    sym = HeapTag(term);
    arity = STArityOfSymbol(sym);

    *info = (_InfoNodeSub *) MemoryAlloc(
       sizeof(_InfoNodeSub) + sizeof(_HeapCell) * arity);

    p = (_IntListNode *) MemoryAlloc(sizeof(_IntListNode));
    p->num = num;
    p->next = NULL;
    (*info)->numslist = p;
    (*info)->ctr = 1;
    newterm = &((*info)->cell);
    HeapInitializeStruct(newterm, sym);

    While arity > 0 Do
      ti = HeapAddr(HeapIthArg(term,arity));
      If HeapIsSymbol(ti) Then
        HeapSetAddr( HeapIthArg(newterm,arity), 
                 InsertSharingSub(ti,num) );
      Else
        HeapSetAddr( HeapIthArg(newterm,arity), ti );
      EndIf
      arity--;
    EndWhile
  EndIf
  Return( newterm );
EndProc



void RemoveNumToSubTerms(_HeapAddr term)
Proc
  Var 
    _HeapAddr ti;
    _IntListNode *p;
    int arity, i;
  _InfoNodeSub *node;
EndVar
  arity = HeapTagArity(term);
  i = arity;
  While i > 0 Do
    ti = HeapAddr(HeapIthArg(term,i));
    If HeapIsSymbol(ti) Then
      RemoveNumToSubTerms( ti );
    EndIf
    i--;
  EndWhile
  node = (_InfoNodeSub*) 
     ((unsigned int)term - sizeof(_IntListNode*) - sizeof(int));
  (node->ctr)--;
  If node->ctr == 0 Then
    STreeDelete(&sub, term);
    While node->numslist Do
      p = node->numslist;
      node->numslist = p->next;
      MemoryFree((_Ptr) p, sizeof(_IntListNode));
    EndWhile
    MemoryFree((_Ptr) node, sizeof(_InfoNodeSub) + sizeof(_HeapCell)*arity);
  EndIf
EndProc




void RemoveSharingSub( _HeapAddr term )
Proc
Var 
  _InfoNodeSub **info; 
  _HeapAddr    newterm;
EndVar

/* ojo: si se viene con el termino que esta en esta estructura, y no
con una copia distinta, entonces nos podemos ahorrar la busqueda */

  info = (_InfoNodeSub **) STreeFind(sub,term);
  newterm = &((*info)->cell);
  RemoveNumToSubTerms(newterm);
EndProc



/*------------------------Rules Output----------------------------*/
void PrintRules()
Proc
Var
  int i;
  _IntListNode *aux;
EndVar

  For i=1;i<=numrules;i++ Do
    If Not DeletedRule(i) Then
      printf("%d.- ",i);
      HeapWriteTerm( rules[i]->lhs );
      If rules[i]->oriented Then
        printf(" --> ");
      Else
        printf(" <-> ");
      EndIf
      HeapWriteTerm( rules[i]->rhs );
      If rules[i]->isold Then printf(" (old)"); EndIf
      printf("   (%d,%d)",rules[i]->parent1,rules[i]->parent2);
      aux = rules[i]->demodlist;
      If aux Then
        printf("   [");
        While aux Do
          printf("%d",aux->num);
          aux = aux->next;
          If aux Then
            printf(",");
          Else
            printf("]");
          EndIf
        EndWhile
      Else
        printf(",[]");
      EndIf
      printf("\n");
    EndIf
  EndFor
  printf("\n\n");
EndProc



void SuperposeTop(_HeapAddr l1, _HeapAddr r1, int p1, 
                  _HeapAddr l2, _HeapAddr r2, int p2)
Proc
Var
  int p;
EndVar
  If p1 > p2 Then p=p2; Else p=p1; EndIf
  If DeletedRule(p1) Or DeletedRule(p2) Then
    Return;
  EndIf
  If p1 != p2 And TermUnify( l1, l2 ) Then
    StoreCP( r1, r2, p1, p2 );
  EndIf
  RestoreRefsUntilEmpty( refStack );
  While numcps > 0 Do
    TreatCP();
  EndWhile
EndProc


void SuperposeSubt1(_HeapAddr s, _HeapAddr l1, _HeapAddr r1, int p1,  
                                 _HeapAddr l2, _HeapAddr r2, int p2)
Proc
Var
  _HeapAddr ti;
  int arity, i;
EndVar

  If DeletedRule(p1) Or DeletedRule(p2) Then
    Return;
  EndIf
  ti = HeapAddr( s );

  If HeapIsSymbol( ti ) And Not HeapIsBlocked(ti) Then
    If TermUnify( ti, l2 ) Then
      HeapSetAddr( s, r2 );
      StoreCP( l1, r1, p1, p2 );
      HeapSetAddr( s, ti );
    EndIf
    RestoreRefsUntilEmpty( refStack );    
    While numcps > 0 Do
      TreatCP();
    EndWhile
    If DeletedRule(p1) Or DeletedRule(p2) Then
      Return;
    EndIf
    arity = HeapTagArity(ti);
    For i=1;i<=arity;i++ Do
      SuperposeSubt1(HeapIthArg(ti,i), l1, r1, p1,   l2, r2, p2);
    EndFor
  EndIf
EndProc


void SuperposeSubt(_HeapAddr l1, _HeapAddr r1, int p1, 
                   _HeapAddr l2, _HeapAddr r2, int p2)
Proc
Var
  int arity, i;
EndVar

  If DeletedRule(p1) Or DeletedRule(p2) Then
    Return;
  EndIf

  If HeapIsSymbol( l1 ) And Not HeapIsBlocked(l1) Then
    arity = HeapTagArity(l1);
    For i=1;i<=arity;i++ Do
      SuperposeSubt1(HeapIthArg(l1,i), l1, r1, p1,    l2, r2, p2);
    EndFor
  EndIf
EndProc



void SuperposeAllWays(_HeapAddr l1, _HeapAddr r1, int p1, 
                      _HeapAddr l2, _HeapAddr r2, int p2)
Proc
  SuperposeTop(  l1, r1, p1, l2, r2, p2 );
  SuperposeSubt( l1, r1, p1, l2, r2, p2 );
  SuperposeSubt( l2, r2, p2, l1, r1, p1 );
EndProc

void CritPairs(int n)
Proc
Var
  _HeapAddr nl,nr;
  int i;
EndVar

  flag4(3,(_HeapAddr)n,rules[n]->lhs, rules[n]->rhs);
  RuleCopyToBank( bankInferences, rules[n]->lhs, rules[n]->rhs, &nl, &nr);
  StackFlush( refStack );

  For i=1;i<=numrules;i++ Do
    If Not DeletedRule(i) And rules[i]->isold Then   /* lhs(n) with lhs(i) */
      SuperposeAllWays( nl, nr, n, rules[i]->lhs, rules[i]->rhs, i );
      If Not rules[i]->oriented Then                 /* l r */
        SuperposeAllWays( nl, nr, n, rules[i]->rhs, rules[i]->lhs, i );
      EndIf
    EndIf
  EndFor

  If Not rules[n]->oriented Then
    For i=1;i<=numrules;i++ Do
      If Not DeletedRule(i) And rules[i]->isold Then   /* r l  */
        SuperposeAllWays( nr, nl, n, rules[i]->lhs, rules[i]->rhs, i );
        If Not rules[i]->oriented Then                 /* r r  */
          SuperposeAllWays( nr, nl, n, rules[i]->rhs, rules[i]->lhs, i );
        EndIf
      EndIf
    EndFor
  EndIf

  flag2(5,(_HeapAddr)n);

  HeapFreeNormalTerm(nl);
  HeapFreeNormalTerm(nr);

EndProc



/*------------------Procedures for size maintenance--------------------*/
void SizeInit()
Proc
Var
  int i;
EndVar
  For i=0;i<MAXRULESIZE;i++ Do
    rulesizes[i]=NULL;
    thmrulesizes[i]=NULL;
  EndFor
  smallestsize=MAXRULESIZE;
  thmsmallestsize=MAXRULESIZE;
EndProc


int GetOldestNewRuleNum()
Proc
  oldestnewrule++;
  While rules[oldestnewrule]->isold Or DeletedRule(oldestnewrule) Do 
    oldestnewrule++;
  EndWhile
  Return(oldestnewrule);
EndProc


int GetSmallestNewRuleNum()
Proc
Var
  _IntListNode *p;
  int n;
EndVar
  n = rulesizes[smallestsize]->num;
  p = rulesizes[smallestsize];
  rulesizes[smallestsize] = (_IntListNode*)rulesizes[smallestsize]->next;
  MemoryFree((_Ptr) p, sizeof(_IntListNode));
  While numnewrules > 0 And rulesizes[smallestsize] == NULL Do
    smallestsize++;
  EndWhile
  If DeletedRule(n) Or rules[n]->isold Then
    n=GetSmallestNewRuleNum();
  EndIf
  Return n;
EndProc



int GetSmallestThmNewRuleNum()
Proc
Var
  _IntListNode *p;
  int n;
EndVar
  n = thmrulesizes[thmsmallestsize]->num;
  p = thmrulesizes[thmsmallestsize];
  thmrulesizes[thmsmallestsize] = (_IntListNode*)thmrulesizes[thmsmallestsize]->next;
  MemoryFree((_Ptr) p, sizeof(_IntListNode));
  While numthmnewrules > 0 And thmrulesizes[thmsmallestsize] == NULL Do
    thmsmallestsize++;
  EndWhile
  If DeletedRule(n) Or rules[n]->isold Then
    n=GetSmallestThmNewRuleNum();
  EndIf
  Return n;
EndProc


void AddNewRuleSize(int size, int rulenum)
Proc
Var
  _IntListNode *p;
EndVar
  p = (_IntListNode *) MemoryAlloc(sizeof(_IntListNode));
  p->next = rulesizes[size];
  p->num = rulenum;
  rulesizes[size] = p;
  If size < smallestsize Then
    smallestsize = size;
  EndIf
EndProc


void AddNewThmRuleSize(int size, int rulenum)
Proc
Var
  _IntListNode *p;
EndVar
  p = (_IntListNode *) MemoryAlloc(sizeof(_IntListNode));
  p->next = thmrulesizes[size];
  p->num = rulenum;
  thmrulesizes[size] = p;
  If size < thmsmallestsize Then
    thmsmallestsize = size;
  EndIf
EndProc


int ComputeRuleSize(_HeapAddr l, _HeapAddr r)
Proc
    Return ComputeTermSize(l)+ComputeTermSize(r);
EndProc


int ComputeTermSize(_HeapAddr t)
Proc
Var
  _HeapAddr auxt, auxs;
  int n,i,arity;
EndVar
  HeapDeref(t, auxt);
  If HeapIsRef(auxt) Then
    Return(1);
  Else
    arity = HeapTagArity(auxt);
    n = arity+1;
    For i = 1; i <= arity; i++ Do
      auxs = HeapIthArg(auxt,i);
      n = n + ComputeTermSize( auxs );
    EndFor
  EndIf
  Return n;
EndProc



_Bool Subsumed(_HeapAddr s, _HeapAddr t)
Proc
Var
  int info;
EndVar
  HeapSetAddr( HeapIthArg(equation,1), s);
  HeapSetAddr( HeapIthArg(equation,2), t);

  info = (int) STreeFirstMatch( ndemod, equation );
  If info Then
    RestoreRefsUntilEmpty( demodRefStack );
    Return(TRUE);
  Else
    RestoreRefsUntilEmpty( demodRefStack );
    HeapSetAddr( HeapIthArg(equation,1), t);
    HeapSetAddr( HeapIthArg(equation,2), s);
    info = (int) STreeFirstMatch( ndemod, equation );
    If info Then
      RestoreRefsUntilEmpty( demodRefStack );
      Return(TRUE);
    Else
      RestoreRefsUntilEmpty( demodRefStack );
      Return(FALSE);
    EndIf
  EndIf
EndProc


void GetProofMembers( int n, _IntListNode **list )
Proc
Var
  _IntListNode *aux, *aux1;
EndVar
  If n<0 Then n= -n; EndIf;
  If n !=0 And rules[n]->parent1 >= 0 Then
    aux = (_IntListNode*) MemoryAlloc( sizeof( _IntListNode ));
    aux->next = *list;
    aux->num = n;
    *list = aux;
    rules[n]->parent1 = -(rules[n]->parent1+1);
    GetProofMembers( -(rules[n]->parent1)-1, list );
    GetProofMembers( rules[n]->parent2, list );
    aux1 = rules[aux->num]->demodlist;
    While aux1 Do
      GetProofMembers( aux1->num, list );
      aux1 = aux1->next;
    EndWhile
  EndIf
EndProc


_IntListNode *GetListDemodSteps()
Proc
Var
  _IntListNode *aux, *aux1; 
EndVar
  aux = NULL;
  While Not StackIsEmpty( demodsteps ) Do
    aux1 = (_IntListNode*) MemoryAlloc( sizeof( _IntListNode ));
    aux1->num = (int) StackTopPop( demodsteps );
    aux1->next = aux;
    aux = aux1;
  EndWhile
  Return(aux);
EndProc

int intcompare(int *i, int *j)
Proc
  Return(*i - *j);
EndProc


void PrintProof(int n)
Proc
Var
  _IntListNode *aux, *aux1;
  int a[1000];
  int i,j;
  FILE *steps;
EndVar
/*  steps = fopen( "steps", "w" );  */
  steps = stdout;
  aux = NULL;
  GetProofMembers( n, &aux );
  i=0;
  While aux Do
    a[i++]=aux->num;
    aux = aux->next;
  EndWhile
  qsort(a,i-1,sizeof(int), (void *) intcompare );
  For j=0;j<i;j++ Do
    If -(rules[a[j]]->parent1)-1 == 0 Then
      fprintf(steps, "[%d=axiom",a[j] );
    ElseIf rules[a[j]]->parent2 == 0 Then
      fprintf(steps, "[%d=demod(%d)",a[j], -(rules[a[j]]->parent1)-1 );
    Else
      fprintf(steps, "[%d=param(%d,%d)",a[j], 
                           -(rules[a[j]]->parent1)-1, rules[a[j]]->parent2);
    EndIf
    aux1 = rules[a[j]]->demodlist;
    If aux1 Then
      fprintf(steps, ",[");
      While aux1 Do
        fprintf(steps, "%d",aux1->num);
        aux1 = aux1->next;
        If aux1 Then
          fprintf(steps, ",");
        Else
          fprintf(steps, "],");
        EndIf
      EndWhile
    Else
      fprintf(steps, ",[],");
    EndIf
    fprintf(steps, "\n\t\t\t");
    HeapWriteTermFile( steps, rules[a[j]]->lhs );
    fprintf(steps, " \t= ");
    HeapWriteTermFile( steps, rules[a[j]]->rhs );
    fprintf(steps, "].\n");
  EndFor
EndProc


void RuleMoveToBank(short int bank, _HeapAddr   l, _HeapAddr   r, 
                                    _HeapAddr *nl, _HeapAddr *nr) 
Proc
Var
  int ctr,h;
EndVar
  ctr=0;
  h = StackHeight( refStack );
  *nl = HeapMoveToBankTerm(&ctr, l, bank);
  *nr = HeapMoveToBankTerm(&ctr, r, bank);
  RestoreRefsUntilHeight( refStack, h );
EndProc  


_HeapAddr CopyCPTerm(_HeapAddr term)
Proc
Var
  _HeapAddr aux,ti;
  _HeapTag tag;
  int i,arity;
  _Bool bank;
EndVar

  bank = FALSE;
  While HeapIsRefOrIntRef(term) And Not HeapIsSelfRef(term) Do
    If VBBankNumber(term) == bankInferences Or
       VBBankNumber(term) == bankDemod Then
      bank=TRUE;
    EndIf
    term = HeapAddr(term);
  EndWhile

  If HeapIsRef(term) Then
    Return(term);
  Else
    tag = HeapTag(term);
    arity = STArityOfSymbol(tag);
    aux = HeapCreateStruct( tag );
    For i=1; i<=arity; i++ Do 
      ti = HeapIthArg(term,i);
      HeapSetAddr( HeapIthArg(aux,i), CopyCPTerm( ti ) );
    EndFor
    If BASIC And (bank Or HeapIsBlocked(term)) Then
      HeapSetBlocked(aux);
    EndIf
    Return(aux);
  EndIf
EndProc


void StoreCP( _HeapAddr s, _HeapAddr t, int p1, int p2 )
Proc
Var 
  _HeapAddr ns,nt,nns,nnt;
EndVar
  flag3(1,s,t);
  numcps++;
  CheckNumCPS;
  critpairs[numcps].p1 = p1;
  critpairs[numcps].p2 = p2;
  ns = CopyCPTerm(s);
  nt = CopyCPTerm(t);
  RuleMoveToBank(bankNormal,ns,nt,&nns,&nnt);
  critpairs[numcps].lhs = nns; 
  critpairs[numcps].rhs = nnt;
EndProc


void FreeCP()
Proc
  HeapFreeTerm(critpairs[numcps].lhs);
  HeapFreeTerm(critpairs[numcps].rhs);
  numcps--;
EndProc


/*------------------- Rewriting -----------------------------*/

_HeapAddr NormalFormlhs(_HeapAddr s)
Proc
  HeapSetAddr( refTermlhs, s );
  ComputeNormalForm( refTermlhs );
  Return( HeapAddr( refTermlhs ));
EndProc


_HeapAddr NormalFormrhs(_HeapAddr s)
Proc
  HeapSetAddr( refTermrhs, s );
  ComputeNormalForm( refTermrhs );
  Return( HeapAddr( refTermrhs ));
EndProc


/* 
   Pre: RHS is built with all refs below a struct pointing
        to another struct or to a variable bank
*/
_HeapAddr CopyRHS(_HeapAddr rhs)
Proc
Var
  _HeapAddr aux;
  _HeapTag tag;
  int i,arity;
EndVar
  If HeapIsRef(rhs) Then  /* it is a BankRef */
    Return(HeapAddr(rhs));
  Else
    tag = HeapTag(rhs);
    arity = STArityOfSymbol(tag);
    aux = rewriteHeap+rewriteHeapFFree;
    rewriteHeapFFree = rewriteHeapFFree+arity+1;
    CheckRewriteHeapFFree;
    HeapInitializeStruct( aux, tag );
    For i=1; i<=arity; i++ Do 
      HeapSetAddr( HeapIthArg(aux,i),
                   CopyRHS( HeapAddr(HeapIthArg(rhs,i)) ) );
    EndFor
    Return(aux);
  EndIf
EndProc


/* PRE: s must be a ref to a term */
void ComputeNormalForm(_HeapAddr s)
Proc
Var
  _HeapAddr add,dref,rhs,lhs,crhs;
  int i,arity,r;
  int info;
EndVar
          
  add = HeapAddr( s );
  HeapDeref( add, dref );
  If HeapIsSymbol(dref) And Not HeapIsNormalForm(dref) Then
    arity = STArityOfSymbol( HeapTag(dref));
    For i=1; i<=arity; i++ Do
      ComputeNormalForm( HeapIthArg(dref,i) );
    EndFor
    info = (int) STreeFirstMatch( demod, dref );
    While info Do
      If info>0 Then
        r = info;
        lhs = rules[r]->lhs;
        rhs = rules[r]->rhs;
      Else
        r = -info;
        rhs = rules[r]->lhs;
        lhs = rules[r]->rhs;
      EndIf
      LPOCacheInit();

      If rules[r]->oriented  Or LPOGr(lhs,rhs,0) Then
        StackPush( demodsteps, (_HeapAddr) info );
        StackPush( rewritestack, add );
        StackPush( rewritestack, s );
        crhs = CopyRHS( rhs );
        If HeapIsSymbol(crhs) And HeapIsBlocked(dref) Then
          HeapSetBlocked(crhs);
        EndIf
        HeapSetAddr( s, crhs );
        RestoreRefsUntilEmpty(demodRefStack);
        info = (int) NULL;
        ComputeNormalForm( s );
      Else
        info = (int) STreeNextMatch();
      EndIf
    EndWhile
    HeapDeref(s,dref);
    If HeapIsSymbol(dref) Then
      HeapSetNormalForm(dref);
    EndIf
  EndIf
EndProc


void BacktrackRewriteStack()
Proc
  While Not StackIsEmpty( rewritestack ) Do
    HeapSetAddr( StackTopPop( rewritestack ), 
                 StackTopPop( rewritestack ) );
  EndWhile
  rewriteHeapFFree=0;
EndProc



void RuleCopyToBank(short int bank, _HeapAddr   l, _HeapAddr   r, 
                                    _HeapAddr *nl, _HeapAddr *nr) 
Proc
Var
  int ctr,h;
EndVar
  *nl = HeapCopyTermProp(l);
  *nr = HeapCopyTermProp(r);
  ctr=0;
  h = StackHeight( refStack );
  *nl = HeapMoveToBankTerm(&ctr, *nl, bank);
  *nr = HeapMoveToBankTerm(&ctr, *nr, bank);
  RestoreRefsUntilHeight( refStack, h );
EndProc  


_Bool IsGround(_HeapAddr t)
Proc
Var
  int arity;
  _Bool ground;
  _HeapAddr term;
EndVar

  HeapDeref(t,term);
  If HeapIsRef(term) Then
    Return(FALSE);
  Else
    ground = TRUE;
    arity = HeapTagArity(term);
    While arity>0 And ground Do
      ground = IsGround( HeapAddr( HeapIthArg(term,arity) ));
      arity--;
    EndWhile
    Return(ground);
  EndIf
EndProc
