%
% save_program( te, p), halt.

% Quintus:
% :- compile(library(unify)).
% message_hook(_,_,_).

p:- 	% write('   <CR> to start checking: '), get0(C),nl,
	read_proof_steps,!,
	setof( P, proof_step(P), Steps1 ),
	extract_step_numbers(Steps1,Steps),
	write('Steps to be checked: '), nl, write(Steps),nl,
	!,
	check_steps( Steps ), halt.

check_steps( [P] ):-
	check(P), !,
        nl,nl,proof_step([P=_,_,Eq]), 
        write('Correct proof of : '), writeeq(Eq), nl,
        actsize(pc,N),
        actsize(eq,N1),
        actsize(pct,N2),
        actsize(eqt,N3),
        write('maximum size cp:                '), write(N),  nl,
        write('maximum size normalized eq:     '), write(N1), nl,
        write('maximum size cp thm:            '), write(N2), nl,
        write('maximum size normalized eq thm: '), write(N3), nl,
	% write('   <CR> to continue'), get0(_),
	nl,nl,nl,nl,nl, !.
check_steps( [P|L] ):- 
	check(P), !,
	check_steps(L),!.
check_steps( [P|_] ):-     
	nl,nl, write('Wrong step! '), nl,
	write('*******************************'), nl, 
	write('*******************************'), nl, 
	write('*******************************'), nl, 
	write('*******************************'), nl, 
	write('*******************************'), nl, 
	write('*******************************'), nl, 
	write('*******************************'), nl, 
	write('*******************************'), nl, 
	write('*******************************'), nl, 
	write('*******************************'), nl, 
	write('*******************************'), nl, 
	write('*******************************'), nl, 
	write('*******************************'), nl, 
	write('*******************************'), nl, 
	write('*******************************'), nl, 
	nl,nl, write('Wrong step! '), nl,
        nl,nl,proof_step([P=S,D,Eq]), 

        write(' computing: '), nl,nl, 
        write(P), write(': '), writeeq(Eq), nl,nl,
        print_step(S),nl,nl, write('  demodulating with: '),nl,nl,
        print_list(D),nl,nl,
	write('   <CR> to continue'), get0(_),nl,nl,nl,nl,nl, !,
	halt.


        

print_step(param(N,M)):-
    write('  by paramodulation of: '),nl,nl,
    proof_step([N=_,_,EN]), write(N), write(': '), writeeq(EN), nl,nl,write('   on '), nl,nl,
    proof_step([M=_,_,EM]), write(M), write(': '), writeeq(EM), nl,!.

print_step(demod(N)):-
    write('  by demod of: '),nl,
    proof_step([N=_,_,EN]), write(N), write(': '), writeeq(EN), nl,!.

print_list([]).
print_list([N|R]):-
    proof_step([N=_,_,EN]), write(N), write(': '), writeeq(EN), nl,nl,nl,
    print_list(R),!.


check(P):-
	proof_step([P=axiom,_,_]),!.
check(P):-
%	write('_______________________________________________'),nl,
%	write('checking step '),write(P),write(...),nl,
	proof_step([P=demod(A),D,Eq]),
	proof_step([A=_, _, Eqa]),
%	write('Trying to get '),
%	writeeq(Eq),nl,
%	write(' from '),nl,
%	write(A), write(': '), writeeq(Eqa),nl,
%	write('demodulating with '),write(D),nl,nl, 
	!,
	demod( 2, Eqa, D, Eq1 ),
	variant_eqs( Eq, Eq1 ),
        keepsize(a=a,Eq),
%	write('correct.'),nl,
%	write('_______________________________________________'),nl,nl,nl,nl,
	!.
check(P):-
%	write('_______________________________________________'),nl,
%	write('checking step '),write(P),write(...),nl,
	proof_step([P=param(A,B),D,Eq]),
	proof_step([A=_, _, Eqa]),
	proof_step([B=_, _, Eqb]), 
%	write('Trying to get '),
%	writeeq(Eq),
%	write(' paramodulating '),nl,
%	write(B), write(': '), writeeq(Eqb),nl,
%	write('on '),nl,
%	write(A), write(': '), writeeq(Eqa),nl,
%	write('and demodulating with '),write(D),nl,nl, 
	!,
        superp( Eqa, Eqb, PC ),
%	write('Critical Pair: '),
%	writeeq(PC),nl,
	demod( 2, PC, D, Eq1 ),
	variant_eqs( Eq, Eq1 ),
        keepsize(PC,Eq),
%	write('correct.'),nl,
%	write('_______________________________________________'),nl,nl,nl,nl,
	!.



        
keepsize(S=T,S1=T1):-
    ( nonvar(S), functor(S,thtop,_); nonvar(T), functor(T,thtop,_)),!,
    size(S,SS),
    size(T,ST), N is SS+ST,
    size(S1,SS1),
    size(T1,ST1), N1 is SS1+ST1,
    storesize(pct,N),
    storesize(eqt,N1),!.
keepsize(S=T,S1=T1):-
    size(S,SS),
    size(T,ST), N is SS+ST,
    size(S1,SS1),
    size(T1,ST1), N1 is SS1+ST1,
    storesize(pc,N),
    storesize(eq,N1),!.


storesize(ID,N):-
    actsize(ID,NA),NA>N,!.
storesize(ID,N):-
    retractall(actsize(ID,_)),
    asserta(actsize(ID,N)),!.


size(X,1):-var(X),!.
size(T,N):-
    functor(T,_,A),
    sizeargs(T,A,N1),
    N is A+1+N1,!.

sizeargs(_,0,0):-!.
sizeargs(T,A,N):-
    arg(A,T,Arg),
    size(Arg,N1),
    A1 is A-1,
    sizeargs(T,A1,N2),
    N is N1 + N2,!.



unify(T1, T2):- T1 == T2, !.
unify(T1, T2):- var(T1), !,
                not_in(T1, T2),
                T1 = T2, !.
unify(T1, T2):- var(T2), !,
                not_in(T2, T1),
                T2 = T1, !.
unify(T1, T2):- T1 =.. [F|Args1],
                T2 =.. [F|Args2], !,
                unifylist(Args1, Args2), !.
unifylist([], []):-!.
unifylist([T1|R1], [T2|R2]):-
                unify(T1, T2),
                unifylist(R1, R2), !.

not_in(X,Y) :-  var(Y),!, X \== Y.
not_in(_,Y) :-  atom(Y),!.
not_in(X,[A1|ARGS]) :- !,
                not_in(X,A1),!,
                not_in(X,ARGS),!.
not_in(X,Y) :-  Y=..[_|ARGY],
                not_in(X,ARGY).


% Quintus:
% match( Left, T ):-
%                subsumes_chk( Left, T ),
%                Left = T,!.


match( Left, T ):-
               match(Left,T,[],L),
               instantiate(L),!.

instantiate([]):-!.
instantiate([ [X,X]|Rest]):-
                instantiate(Rest),!.


match(Left,T,L,L):-
                var(Left),
                in_subst([Left,B],L),!,
                B==T,!.
match(Left,T,L, [[Left,T]|L] ):-
                var(Left),!.
match(Left,T,L1,L2) :-
                nonvar(T),
                functor(Left,F,N),!,
                functor(T,F,N),!,
                Left =.. [_|ARGLeft],!,
                T =.. [_|ARGT],!,
                match_list(ARGLeft,ARGT,L1,L2).

match_list( [],[],L,L ):- !.
match_list([A1|ARGS1],[A2|ARGS2],L1,L2 ):-
                match(A1,A2,L1,L3),!,
                match_list(ARGS1,ARGS2,L3,L2),!.

in_subst( [L,T], [[X,T]|_] ):- L == X,!.
in_subst( X, [_|R] ):- in_subst( X,R ).

%------------------------------------------------------------------------

% otherwise sharedsamesteps may not terminate
orderable(L,R):-
    subterm(L1,R,_,_),
    L1 == L, !,fail.
orderable(_,_):-!.


/*
sharedsamesteps( _,NE, _, NE ).
sharedsamesteps( L,E,R,E1 ):-
	subterm(Sub,E,NE,R),
	L == Sub,
        sharedsamesteps( L,NE,R,E1 ).

*/

sharedsamesteps( L,E,R,E1 ):-
    count_occs(L,E,N), !, N<5, !,
    subset(N,S),
    replace_ocs(1,_,S,L,R,E,E1).


q:- sharedsamesteps( a,f(a,a,a),b,E ),write(E),nl,fail.

subset(0,[]):-!.
subset(N,S1):-
    N1 is N-1,
    subset(N1,S),
    append(S,[N],S1).
subset(N,S):-
    N1 is N-1,
    subset(N1,S).


count_occs(_,E, 0):- var(E),!.
count_occs(_,[],0):-!.
count_occs(L,E, 1):- L==E,!.
count_occs(L,[E|R],N):-
    count_occs(L,E,N1),
    count_occs(L,R,N2),
    N is N1 + N2,!.
count_occs(L,E,N):-
    E =.. [_|Args],
    count_occs(L,Args,N),!.


replace_ocs(_,_,[],_,_,E,E):-!.
replace_ocs(N,N,_,_,_,E,E):- var(E),!.
replace_ocs(N,N,_,_,_,[],[]):-!.
replace_ocs(N,N1,S,L,R,E,R):- L==E, member(N,S), N1 is N+1,!.
replace_ocs(N,N1,_,L,_,E,E):- L==E, N1 is N+1,!.
replace_ocs(N,N1,S,L,R,[E1|A1],[E2|A2]):-
    replace_ocs(N, N2,S,L,R,E1,E2),
    replace_ocs(N2,N1,S,L,R,A1,A2),!.
replace_ocs(N,N1,S,L,R,E1,E2):-
    E1 =.. [F|Args1],
    replace_ocs(N,N1,S,L,R,Args1,Args2),
    E2 =.. [F|Args2],!.



demod( _, E, [], E ):-!.
demod( _, _, [], _ ):-!,fail.
demod( N, E, [D|Rest], E1 ):-    D>0,!,
        proof_step([D=_, _, L=R]),
	subterm(Sub,E,NE,R),
        nonvar(L),
        match( L, Sub ),
	N1 is N + 1,
        sharedsamesteps( L,NE,R,NNE ),
%	spaces(N), write('demod w/ '), write(D), 
%	write(': '), writeeq(NNE),nl,
	demod( N1, NNE, Rest, E1 ).
demod( N, E, [D|Rest], E1 ):-
        D1 is -D,
        proof_step([D1=_, _, R=L]),
	subterm(Sub,E,NE,R),
        nonvar(L),
        match( L, Sub ),
	N1 is N + 1,
        sharedsamesteps( L,NE,R,NNE ),
%	spaces(N), write('demod w/ '), write(D), 
%	write(': '), writeeq(NNE),nl,
	demod( N1, NNE, Rest, E1 ).



% Quintus:
%variant_eqs( S=T, E ):-
%	subsumes_chk( S=T, E ),
%	subsumes_chk( E, S=T ),!.
%variant_eqs( T=S, E ):-
%	subsumes_chk( S=T, E ),
%	subsumes_chk( E, S=T ),!.


variant_eqs( S=T, E ):-
	match(S=T,E,[],_),
	match(E,S=T,[],_),!.
variant_eqs( T=S, E ):-
	match(S=T,E,[],_),
	match(E,S=T,[],_),!.

superp(L1=R1,L2=R2,CP):-
	superp1(L1=R1,L2=R2,CP).
superp(L1=R1,R2=L2,CP):-
	superp1(L1=R1,L2=R2,CP).
superp(R1=L1,L2=R2,CP):-
	superp1(L1=R1,L2=R2,CP).
superp(R1=L1,R2=L2,CP):-
	superp1(L1=R1,L2=R2,CP).

superp1(L1=R1,L2=R2,CP=R1):-
	subterm(Sub,L1,CP,R2),
	nonvar(Sub),
	unify(Sub,L2).

subterm(T,T,Vsubst,Vsubst).
subterm(S,T,NT,Vsubst) :-
	nonvar(T),
	T =.. [F|Args1],
	subterm_list(S,Args1,Args2,Vsubst),
	NT =.. [F|Args2].
subterm_list(S,[T|Rest],[NT|Rest],Vsubst):-
	nonvar(T),
	subterm(S,T,NT,Vsubst).
subterm_list(S,[T|Rest1],[T|Rest2],Vsubst):-
	subterm_list(S,Rest1,Rest2,Vsubst).


extract_step_numbers([],[]).
extract_step_numbers([[P=_|_]|R],[P|R1]):-
	extract_step_numbers(R,R1).


read_proof_steps:- 
	retractall(proof_step(_)),
	retractall(actsize(_,_)),
        asserta(actsize(pc,0)),
        asserta(actsize(eq,0)),
        asserta(actsize(pct,0)),
        asserta(actsize(eqt,0)),
%	write('File:'), read(F), see(F),
	see(steps),
	repeat, read(L), the_end(L), seen.


the_end('end_of_file'):-!.
the_end(L):- 
%	writeeq(L), nl, 
	assertz(proof_step(L)), fail.


spaces(0):-!.
spaces(N):-
	write(' '), N1 is N-1, spaces(N1).

writeeq(Term):- \+(\+(print_term1(Term))), !.
print_term1(Term):-
	numbervars( Term, 0, _), !, 
	write_term(Term, [numbervars(true)]), !.





member(X,[X|_]).
member(X,[_|L]):-member(X,L).



append([],L,L).
append([X|L1],L2,[X|L3]):- append(L1,L2,L3).





:-p.
